/*
 * =========================================================================
 * Copyright (C) 1997 - 1998 by Visual Numerics, Inc. All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software is freely
 * granted by Visual Numerics, Inc., provided that the copyright notice
 * above and the following warranty disclaimer are preserved in human
 * readable form.
 *
 * Because this software is licenses free of charge, it is provided
 * "AS IS", with NO WARRANTY.  TO THE EXTENT PERMITTED BY LAW, VNI
 * DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO ITS PERFORMANCE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * VNI WILL NOT BE LIABLE FOR ANY DAMAGES WHATSOEVER ARISING OUT OF THE USE
 * OF OR INABILITY TO USE THIS SOFTWARE, INCLUDING BUT NOT LIMITED TO DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, PUNITIVE, AND EXEMPLARY DAMAGES, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES. 
 *
 * =========================================================================
 */

package com.imsl.test.math;
import com.imsl.math.Complex;
import com.imsl.math.Sfun;

class TestComplex extends Test
{
	final private double radius[] = {0.5, 1.0e300};
	final private double angles[] = {
		 0.00001*PI,
		 0.24999*PI,
		 0.02501*PI,
		 0.49999*PI,
		-0.49999*PI,
		-0.25001*PI,
		-0.24999*PI,
		-0.00001*PI,
		 2.99999*PI,
		 3.00001*PI
	};

	void negative()
	{
		Complex z;
		z = new Complex(3,4);
		check("negative(3+4i)", Complex.negative(z), new Complex(-3,-4));
	}

	void conjugate()
	{
		Complex z;
		z = new Complex(3,4);
		check("conjugate(3+4i)", Complex.conjugate(z), new Complex(3,-4));
	}

	void plus()
	{
		Complex w, z;
		w = new Complex(2,1);
		z = new Complex(3,5);
		check("add(2+i,3+5i)", Complex.plus(w,z), new Complex(5,6));
	}

	void minus()
	{
		Complex w, z;
		w = new Complex(3,5);
		z = new Complex(2,1);
		check("subtract(2+i,3+5i)", Complex.minus(w,z), new Complex(1,4));
	}

	void timesCC()
	{
		Complex w, z;
		w = new Complex(3,5);
		z = new Complex(2,1);
		
		check("times(3+5i,2+i)",
			Complex.times(w,z),
			new Complex(1,13));
		
		check("times(2+i,3+5i)",
			w.times(z),
			new Complex(1,13));

		w = new Complex(3,5);
		z = new Complex(2,1);

		check("times(+inf+i,3+i*5)",
			Complex.times(new Complex(POSINF,1),w),
			new Complex(POSINF,POSINF));
		
		check(
			"times(+inf+i,NaN+i*5)",
			Complex.times(new Complex(POSINF,1),new Complex(NaN,5)),
			new Complex(NaN,POSINF));
	}

	void timesDC()
	{
		Complex z;
		double x = 3;
		z = new Complex(2,1);
		
		check("times(3,2+i)",
			Complex.times(x,z),
			new Complex(6,3));
	}

	void timesCD()
	{
		Complex z;
		double x = 3;
		z = new Complex(2,1);
		
		check("times(2+i,3)",
			Complex.times(z,x),
			new Complex(6,3));
	}

	void divideCC()
	{
		Complex w = new Complex(3,5);
		Complex z = new Complex(2,1);
		Complex t = new Complex(POSINF,2);

		
		check("over(3+5i,2+i)",
			Complex.over(w,z),
			new Complex(2.2,1.4));

		check("over(2+i,+inf+2i)",
			Complex.over(z,t),
			new Complex(0,0));

		check("over(2+i,+2+i*inf)",
			Complex.over(z,new Complex(2,POSINF)),
			new Complex(0,NEGZERO));

		check("over(+inf+2i,2+i)",
			Complex.over(t,z),
			new Complex(POSINF,NEGINF));
	}

	void divideCD()
	{
		Complex z = new Complex(2,1);
		double x = 3;
		double t = 1.0/3.0;

		check("over(2+i,3)",
			Complex.over(z,x),
			new Complex(2*t,t));
		
		check("(2+i).over(3)",
			Complex.over(z,x),
			new Complex(2*t,t));
	}

	void abs()
	{
		check("abs(3+i*4)",
			Complex.abs(new Complex(3,4)),
			5.0);

		check("abs(+inf+i*4)",
			Complex.abs(new Complex(POSINF,4)),
			POSINF);
		check("abs(-inf+i*4)",
			Complex.abs(new Complex(NEGINF,4)),
			POSINF);
		check("abs(+3+i*inf)",
			Complex.abs(new Complex(3,POSINF)),
			POSINF);
		check("abs(NaN+i*4)",
			Complex.abs(new Complex(NaN,4)),
			NaN);
		check("abs(+3+i*NaN)",
			Complex.abs(new Complex(3,NaN)),
			NaN);
		check("abs(+inf+i*NaN)",
			Complex.abs(new Complex(POSINF,NaN)),
			POSINF);
		check("abs(+NaN-i*inf)",
			Complex.abs(new Complex(NaN,NEGINF)),
			POSINF);
		check("abs(NaN+i*NaN)",
			Complex.abs(new Complex(NaN,NaN)),
			NaN);
	}

	void argument()
	{
		Complex z;
		z = new Complex(3,4);
		check("argument1", Complex.argument(z), 0.927295218001612);
	}

	void sqrt()
	{
		check("sqrt(3+4i)",
			Complex.sqrt(new Complex(3,4)),
			new Complex(2,1));
		check("sqrt(30000+40000i)",
			Complex.sqrt(new Complex(30000,40000)),
			new Complex(200,100));

		// Special cases
		check("sqrt(0)",
			Complex.sqrt(new Complex(0,0)),
			new Complex(0,0));
		check("sqrt(-inf+i)",
			Complex.sqrt(new Complex(NEGINF,1)),
			new Complex(0,POSINF));
		check("sqrt(+inf+i)",
			Complex.sqrt(new Complex(POSINF,1)),
			new Complex(POSINF,0));
		check("sqrt(1+i*inf)",
			Complex.sqrt(new Complex(1,POSINF)),
			new Complex(POSINF,POSINF));
		check("sqrt(NaN+i*inf)",
			Complex.sqrt(new Complex(NaN,POSINF)),
			new Complex(POSINF,POSINF));
		check("sqrt(-inf+i*NaN)",
			Complex.sqrt(new Complex(NEGINF,NaN)),
			new Complex(NaN,POSINF));
		check("sqrt(+inf+i*NaN)",
			Complex.sqrt(new Complex(POSINF,NaN)),
			new Complex(POSINF,NaN));
		check("sqrt(1+i*NaN)",
			Complex.sqrt(new Complex(1,NaN)),
			new Complex(NaN,NaN));
		check("sqrt(NaN+i)",
			Complex.sqrt(new Complex(NaN,1)),
			new Complex(NaN,NaN));
		check("sqrt(NaN+i*NaN)",
			Complex.sqrt(new Complex(NaN,NaN)),
			new Complex(NaN,NaN));
		
		// Special cases conjugate
		check("sqrt(-inf-i)",
			Complex.sqrt(new Complex(NEGINF,-1)),
			new Complex(0,NEGINF));
		check("sqrt(+inf-i)",
			Complex.sqrt(new Complex(POSINF,-1)),
			new Complex(POSINF,NEGZERO));
		check("sqrt(1-i*inf)",
			Complex.sqrt(new Complex(1,NEGINF)),
			new Complex(POSINF,NEGINF));	
		check("sqrt(-inf-i*NaN)",
			Complex.sqrt(new Complex(NEGINF,NaN)),
			new Complex(NaN,POSINF));
		check("sqrt(+inf-i*NaN)",
			Complex.sqrt(new Complex(POSINF,NaN)),
			new Complex(POSINF,NaN));
		check("sqrt(1-i*NaN)",
			Complex.sqrt(new Complex(1,NaN)),
			new Complex(NaN,NaN));
		check("sqrt(NaN-i)",
			Complex.sqrt(new Complex(NaN,-1)),
			new Complex(NaN,NaN));
		check("sqrt(NaN-i*NaN)",
			Complex.sqrt(new Complex(NaN,-NaN)),
			new Complex(NaN,NaN));

		// Check near possible branch cuts
		for (int j = 0;  j < radius.length;  j++) {
			double	sqrtr = Math.sqrt(radius[j]);
			for (int k = 0;  k < angles.length;  k++) {
				double p = (angles[k] % (2*PI));
				if (p > PI)  p -= 2*PI;
				check("sqrt("+radius[j]+"*exp("+angles[k]/PI+"*pi*i))",
					Complex.sqrt(new Complex(radius[j]*Math.cos(angles[k]),radius[j]*Math.sin(angles[k]))),
					new Complex(sqrtr*Math.cos(p/2),sqrtr*Math.sin(p/2)));
			}
		}
	}

	void exp()
	{
		check("exp(2+i)", Complex.exp(new Complex(2,1)),  new Complex(3.99232404844127, 6.21767631236797));
		check("exp(2-i)", Complex.exp(new Complex(2,-1)), new Complex(3.99232404844127,-6.21767631236797));

		// Special cases
		check("exp(0+i0)",
			Complex.exp(new Complex(0,0)),
			new Complex(1,0));
		check("exp(-0+i0)",
			Complex.exp(new Complex(NEGZERO,0)),
			new Complex(1,0));
		check("exp(+inf+i0)",
			Complex.exp(new Complex(POSINF,0)),
			new Complex(POSINF,0));
		check("exp(-inf+i*inf)",
			Complex.exp(new Complex(NEGINF,POSINF)),
			new Complex(0,0));
		check("exp(1+i*inf)",
			Complex.exp(new Complex(1,POSINF)),
			new Complex(NaN,NaN));
		check("exp(-inf+2*i)",
			Complex.exp(new Complex(NEGINF,2)),
			new Complex(Math.cos(2)/POSINF,Math.sin(2)/POSINF)); //???
		check("exp(+inf+2*i)",
			Complex.exp(new Complex(POSINF,2)),
			new Complex(POSINF*Math.cos(2),POSINF*Math.sin(2)));
		check("exp(-inf+i*NaN)",
			Complex.exp(new Complex(NEGINF,NaN)),
			new Complex(0,0));
		check("exp(+inf+i*NaN)",
			Complex.exp(new Complex(POSINF,NaN)),
			new Complex(POSINF,NaN));
		check("exp(NaN+i*0)",
			Complex.exp(new Complex(NaN,0)),
			new Complex(NaN,0));
		check("exp(NaN+i*2)",
			Complex.exp(new Complex(NaN,2)),
			new Complex(NaN,NaN));
		check("exp(2+i*NaN)",
			Complex.exp(new Complex(2,NaN)),
			new Complex(NaN,NaN));
		check("exp(NaN+i*NaN)",
			Complex.exp(new Complex(NaN,NaN)),
			new Complex(NaN,NaN));

		// Special cases conjugate
		check("exp(0-i0)",
			Complex.exp(new Complex(0,NEGZERO)),
			new Complex(1,NEGZERO));
		check("exp(-0-i0)",
			Complex.exp(new Complex(NEGZERO,NEGZERO)),
			new Complex(1,NEGZERO));
		check("exp(+inf-i0)",
			Complex.exp(new Complex(POSINF,NEGZERO)),
			new Complex(POSINF,NEGZERO));
		check("exp(-inf-i*inf)",
			Complex.exp(new Complex(NEGINF,NEGINF)),
			new Complex(0,0));
		check("exp(1-i*inf)",
			Complex.exp(new Complex(1,NEGINF)),
			new Complex(NaN,NaN));
		check("exp(-inf-2*i)",
			Complex.exp(new Complex(NEGINF,-2)),
			new Complex(0*Math.cos(2),NEGZERO*Math.sin(2))); //???
		check("exp(+inf-2*i)",
			Complex.exp(new Complex(POSINF,-2)),
			new Complex(POSINF*Math.cos(2),NEGINF*Math.sin(2)));
		check("exp(-inf+i*NaN)",
			Complex.exp(new Complex(NEGINF,NaN)),
			new Complex(0,0));
		check("exp(+inf+i*NaN)",
			Complex.exp(new Complex(POSINF,NaN)),
			new Complex(POSINF,NaN));
		check("exp(NaN-i*0)",
			Complex.exp(new Complex(NaN,NEGZERO)),
			new Complex(NaN,NEGZERO));
		check("exp(NaN-i*2)",
			Complex.exp(new Complex(NaN,-2)),
			new Complex(NaN,NaN));

		// Check near possible branch cuts
		for (int j = 0;  j < radius.length;  j++) {
			double	r = Math.min(10,radius[j]);	
			for (int k = 0;  k < angles.length;  k++) {
				double x = r*Math.cos(angles[k]);
				double y = r*Math.sin(angles[k]);
				double expx = Math.exp(x);
				check("exp("+r+"*exp("+angles[k]/PI+"*pi*i))",
					Complex.exp(new Complex(x,y)),
					new Complex(expx*Math.cos(y),expx*Math.sin(y)));
			}
		}
	}

	void log()
	{
		check("log(2+i)", Complex.log(new Complex(2,1)), new Complex(0.80471895621705,0.463647609000806));

		// Special cases
		check("log(-0+i0)",
			Complex.log(new Complex(NEGZERO,0)),
			new Complex(NEGINF,PI));
		check("log(+0+i0)",
			Complex.log(new Complex(0,0)),
			new Complex(NEGINF,0));
		check("log(-inf+i*inf)",
			Complex.log(new Complex(NEGINF,POSINF)),
			new Complex(POSINF,0.75*PI));
		check("log(+inf+i*inf)",
			Complex.log(new Complex(POSINF,POSINF)),
			new Complex(POSINF,0.25*PI));
		check("log(1+i*inf)",
			Complex.log(new Complex(1,POSINF)),
			new Complex(POSINF,0.50*PI));
		check("log(-inf+i)",
			Complex.log(new Complex(NEGINF,1)),
			new Complex(POSINF,PI));
		check("log(+inf+i*inf)",
			Complex.log(new Complex(POSINF,1)),
			new Complex(POSINF,0));
		check("log(+inf+i*NaN)",
			Complex.log(new Complex(POSINF,NaN)),
			new Complex(POSINF,NaN));
		check("log(-inf+i*NaN)",
			Complex.log(new Complex(POSINF,NaN)),
			new Complex(POSINF,NaN));
		check("log(NaN+i*inf)",
			Complex.log(new Complex(NaN,POSINF)),
			new Complex(POSINF,NaN));
		check("log(1+i*NaN)",
			Complex.log(new Complex(1,NaN)),
			new Complex(NaN,NaN));
		check("log(NaN+i)",
			Complex.log(new Complex(NaN,1)),
			new Complex(NaN,NaN));
		check("log(NaN+i*NaN)",
			Complex.log(new Complex(NaN,NaN)),
			new Complex(NaN,NaN));

		// Special cases conjugates
		check("log(-0-i0)",
			Complex.log(new Complex(NEGZERO,NEGZERO)),
			new Complex(NEGINF,-PI));
		check("log(+0-i0)",
			Complex.log(new Complex(0,NEGZERO)),
			new Complex(NEGINF,NEGZERO));
		check("log(-inf-i*inf)",
			Complex.log(new Complex(NEGINF,NEGINF)),
			new Complex(POSINF,-0.75*PI));
		check("log(+inf-i*inf)",
			Complex.log(new Complex(POSINF,NEGINF)),
			new Complex(POSINF,-0.25*PI));
		check("log(1-i*inf)",
			Complex.log(new Complex(1,NEGINF)),
			new Complex(POSINF,-0.50*PI));
		check("log(-inf-i)",
			Complex.log(new Complex(NEGINF,-1)),
			new Complex(POSINF,-PI));
		check("log(+inf-i*inf)",
			Complex.log(new Complex(POSINF,-1)),
			new Complex(POSINF,NEGZERO));
		check("log(NaN-i*inf)",
			Complex.log(new Complex(NaN,NEGINF)),
			new Complex(POSINF,NaN));
		check("log(NaN-i)",
			Complex.log(new Complex(NaN,-1)),
			new Complex(NaN,NaN));

		// Check near possible branch cuts
		for (int j = 0;  j < radius.length;  j++) {
			double	logr = Math.log(radius[j]);
			for (int k = 0;  k < angles.length;  k++) {
				double p = (angles[k] % (2*PI));
				if (p > PI)  p -= 2*PI;
				check("log("+radius[j]+"*exp("+angles[k]/PI+"*pi*i))",
					Complex.log(new Complex(radius[j]*Math.cos(angles[k]),radius[j]*Math.sin(angles[k]))),
					new Complex(logr,p));
			}
		}

		for (int k = 0;  k < REPEAT;  k++) {
			Complex z = new Complex(random(),random());
			check(k+"log series "+z, Complex.log(Complex.exp(z)), z);
		}
	}

	void sin()
	{
		check("sin(2+i)",
			Complex.sin(new Complex(2,1)),
			new Complex(1.40311925062204,-0.489056259041294));
	}

	void cos()
	{
		check("cos(2+i)",
			Complex.cos(new Complex(2,1)),
			new Complex(-0.64214812471552,-1.06860742138278));
	}

	void tan()
	{
		check("tan(2+i)",
			Complex.tan(new Complex(2,1)),
			new Complex(-0.243458201185724,1.16673625724092));
	}


	void asin()
	{
		Complex z;

		for (int k = 0;  k < REPEAT;  k++) {
			z = new Complex(0.1,-2+4*random());
			check("sin(asin("+z+"))",
				Complex.sin(Complex.asin(z)),
				z);
		}
			
		for (int k = 0;  k < REPEAT;  k++) {
			z = new Complex(-2+4*random(),0.1);
			check("sin(asin("+z+"))",
				Complex.sin(Complex.asin(z)),
				z);
		}

		for (int k = 0;  k < REPEAT;  k++) {
			z = new Complex(-2*4*random(),-2+4*random());
			check("sin(asin("+z+"))",
				Complex.sin(Complex.asin(z)),
				z);
		}

		for (int k = 0;  k < REPEAT;  k++) {
			z = new Complex(PI/2-0.05*random(),0.05*random());
			check("sin(asin("+z+"))",
				Complex.sin(Complex.asin(z)),
				z);
		}

		check("asin(+inf+i*2)",
			Complex.asin(new Complex(POSINF,2)),
			new Complex(PI/2,POSINF));
	}


	void acos()
	{
		Complex z;

		for (int k = 0;  k < REPEAT;  k++) {
			z = new Complex(0.1,-2+4*random());
			check(
				"cos(acos("+z+"))",
				Complex.cos(Complex.acos(z)),
				z);
		}
			
		for (int k = 0;  k < REPEAT;  k++) {
			z = new Complex(-2+4*random(),0.1);
			check(
				"cos(acos("+z+"))",
				Complex.cos(Complex.acos(z)),
				z);
		}
			
		for (int k = 0;  k < REPEAT;  k++) {
			z = new Complex(-2*4*random(),-2+4*random());
			check(
				"cos(acos("+z+"))",
				Complex.cos(Complex.acos(z)),
				z);
		}

		for (int k = 0;  k < REPEAT;  k++) {
			z = new Complex(PI/2-0.05*random(),0.05*random());
			check(
				"cos(acos("+z+"))",
				Complex.cos(Complex.acos(z)),
				z);
		}

		// Special cases
		check("acos(+0+i*0)",
			Complex.acos(new Complex(0,0)),
			new Complex(PI/2,NEGZERO));
		check("acos(-0+i*0)",
			Complex.acos(new Complex(NEGZERO,0)),
			new Complex(PI/2,NEGZERO));
		check("acos(-inf+i*inf)",
			Complex.acos(new Complex(NEGINF,POSINF)),
			new Complex(0.75*PI,NEGINF));
		check("acos(+inf+i*inf)",
			Complex.acos(new Complex(POSINF,POSINF)),
			new Complex(0.25*PI,NEGINF));
		check("acos(2+i*inf)",
			Complex.acos(new Complex(2,POSINF)),
			new Complex(PI/2,NEGINF));
		check("acos(-inf+i*2)",
			Complex.acos(new Complex(NEGINF,2)),
			new Complex(PI,NEGINF));
		check("acos(+inf+i*2)",
			Complex.acos(new Complex(POSINF,2)),
			new Complex(0,NEGINF));
		check("acos(+0+i*NaN)",
			Complex.acos(new Complex(0,NaN)),
			new Complex(PI/2,NaN));
		check("acos(NaN+i*inf)",
			Complex.acos(new Complex(NaN,POSINF)),
			new Complex(NaN,NEGINF));
		check("acos(2+i*NaN)",
			Complex.acos(new Complex(2,NaN)),
			new Complex(NaN,NaN));
		check("acos(NaN+i*2)",
			Complex.acos(new Complex(NaN,2)),
			new Complex(NaN,NaN));
		check("acos(NaN+i*NaN)",
			Complex.acos(new Complex(NaN,NaN)),
			new Complex(NaN,NaN));

	// Special cases (conjugate)
		check("acos(+0-i*0)",
			Complex.acos(new Complex(0,NEGZERO)),
			new Complex(PI/2,0));
		check("acos(-0-i*0)",
			Complex.acos(new Complex(NEGZERO,NEGZERO)),
			new Complex(PI/2,0));
		check("acos(-inf-i*inf)",
			Complex.acos(new Complex(NEGINF,NEGINF)),
			new Complex(0.75*PI,POSINF));
		check("acos(+inf-i*inf)",
			Complex.acos(new Complex(POSINF,NEGINF)),
			new Complex(0.25*PI,POSINF));
		check("acos(2-i*inf)",
			Complex.acos(new Complex(2,NEGINF)),
			new Complex(PI/2,POSINF));
		check("acos(-inf-i*2)",
			Complex.acos(new Complex(NEGINF,-2)),
			new Complex(PI,POSINF));
		check("acos(+inf-i*2)",
			Complex.acos(new Complex(POSINF,-2)),
			new Complex(0,POSINF));
		check("acos(NaN-i*inf)",
			Complex.acos(new Complex(NaN,NEGINF)),
			new Complex(NaN,POSINF));
		check("acos(NaN-i*2)",
			Complex.acos(new Complex(NaN,-2)),
			new Complex(NaN,NaN));
	}

	void atan()
	{
		Complex ans, z;
		
		ans = new Complex(.2,.1);
		check("atan(tan(0.2+0.1i))",
			Complex.atan(Complex.tan(ans)),
			ans);
		
		ans = new Complex(1,2);
		check("atan(tan(1+2i))",
			Complex.atan(Complex.tan(ans)),
			ans);

		ans = new Complex(1.0e-20,4);
		check("atan(tan(1.0e-20+4i))",
			Complex.tan(Complex.atan(ans)),
			ans);

		ans = new Complex(0,1.0e-20);
		check("atan(tan(0+1.0e-20i))",
			Complex.tan(Complex.atan(ans)),
			ans);

		check("atan(i)",
			Complex.atan(new Complex(0,1)),
			new Complex(0,POSINF));
		check("tan(0+i*inf)",
			Complex.tan(new Complex(0,POSINF)),
			new Complex(0,1));

		for (int k = 0;  k < REPEAT;  k++) {
			z = new Complex(0.1*random(),0.1*random());
			check(k+"atan(tan("+z+"))",
				Complex.atan(Complex.tan(z)),
				z);
		}
	}

	void sinh()
	{
		check("sinh(2)",
			Complex.sinh(new Complex(2)),
			new Complex(Sfun.sinh(2)));

		check("sinh(0+i*PI/2)",
			Complex.sinh(new Complex(0,PI/2)),
			new Complex(0,1));
		check("sinh(0+i*3*PI/2)",
			Complex.sinh(new Complex(0,3*PI/2)),
			new Complex(NEGZERO,-1));

		// Special cases
		check("sinh(+0+i*0)",
			Complex.sinh(new Complex(0,0)),
			new Complex(0,0));
		check("sinh(+0+i*inf)",
			Complex.sinh(new Complex(0,POSINF)),
			new Complex(0,NaN));
		check("sinh(+inf+i*0)",
			Complex.sinh(new Complex(POSINF,0)),
			new Complex(POSINF,0));
		check("sinh(+inf+i*inf)",
			Complex.sinh(new Complex(POSINF,POSINF)),
			new Complex(POSINF,NaN));	// can be +-inf + iNaN
		check("sinh(+inf+i)",
			Complex.sinh(new Complex(POSINF,1)),
			new Complex(POSINF*Math.cos(1),POSINF*Math.sin(1)));
		check("sinh(+inf+i*2)",
			Complex.sinh(new Complex(POSINF,2)),
			new Complex(POSINF*Math.cos(2),POSINF*Math.sin(2)));
		check("sinh(+1+i*inf)",
			Complex.sinh(new Complex(1,POSINF)),
			new Complex(NaN,NaN));
		check("sinh(NaN+i*0)",
			Complex.sinh(new Complex(NaN,0)),
			new Complex(NaN,0));
		check("sinh(NaN+i)",
			Complex.sinh(new Complex(NaN,1)),
			new Complex(NaN,NaN));
		check("sinh(NaN+i*NaN)",
			Complex.sinh(new Complex(NaN,NaN)),
			new Complex(NaN,NaN));

		// Special cases conjugate
		check("sinh(+0-i*0)",
			Complex.sinh(new Complex(0,NEGZERO)),
			new Complex(0,0));
		check("sinh(+0-i*inf)",
			Complex.sinh(new Complex(0,NEGINF)),
			new Complex(0,NaN));
		check("sinh(+inf-i*0)",
			Complex.sinh(new Complex(POSINF,NEGZERO)),
			new Complex(POSINF,0));
		check("sinh(+inf-i*inf)",
			Complex.sinh(new Complex(POSINF,NEGINF)),
			new Complex(NEGINF,NaN));	// can be +-inf + iNaN
		check("sinh(+inf-i)",
			Complex.sinh(new Complex(POSINF,-1)),
			new Complex(POSINF*Math.cos(-1),POSINF*Math.sin(-1)));
		check("sinh(+inf-i*2)",
			Complex.sinh(new Complex(POSINF,-2)),
			new Complex(POSINF*Math.cos(-2),POSINF*Math.sin(-2)));
		check("sinh(+1-i*inf)",
			Complex.sinh(new Complex(1,NEGINF)),
			new Complex(NaN,NaN));
		check("sinh(NaN-i*0)",
			Complex.sinh(new Complex(NaN,NEGZERO)),
			new Complex(NaN,0));
		check("sinh(NaN-i)",
			Complex.sinh(new Complex(NaN,-1)),
			new Complex(NaN,NaN));

		// Special cases (oddness)
		check("sinh(-inf-i*0)",
			Complex.sinh(new Complex(NEGINF,NEGZERO)),
			new Complex(NEGINF,0));
		check("sinh(-inf-i*inf)",
			Complex.sinh(new Complex(NEGINF,NEGINF)),
			new Complex(NEGINF,NaN));	// can be +-inf + iNaN
		check("sinh(-inf-i)",
			Complex.sinh(new Complex(NEGINF,-1)),
			new Complex(NEGINF*Math.cos(-1),NEGINF*Math.sin(-1)));
		check("sinh(-inf-i*2)",
			Complex.sinh(new Complex(NEGINF,-2)),
			new Complex(NEGINF*Math.cos(-2),NEGINF*Math.sin(-2)));
		check("sinh(-1-i*inf)",
			Complex.sinh(new Complex(-1,NEGINF)),
			new Complex(NaN,NaN));
		check("sinh(NaN-i*0)",
			Complex.sinh(new Complex(NaN,NEGZERO)),
			new Complex(NaN,0));
		check("sinh(NaN-i)",
			Complex.sinh(new Complex(NaN,-1)),
			new Complex(NaN,NaN));
	}

	void cosh()
	{
		check("cosh(2)", Complex.cosh(new Complex(2)), new Complex(Sfun.cosh(2)));

		// Special cases
		check("cosh(+0+i*0)",
			Complex.cosh(new Complex(0,0)),
			new Complex(1,0));
		check("cosh(+0+i*inf)",
			Complex.cosh(new Complex(0,POSINF)),
			new Complex(NaN,0));
		check("cosh(+inf+i*0)",
			Complex.cosh(new Complex(POSINF,0)),
			new Complex(POSINF,0));
		check("cosh(+inf+i*inf)",
			Complex.cosh(new Complex(POSINF,POSINF)),
			new Complex(POSINF,NaN));	// can be +-inf + iNaN
		check("cosh(+inf+i)",
			Complex.cosh(new Complex(POSINF,1)),
			new Complex(POSINF*Math.cos(1),POSINF*Math.sin(1)));
		check("cosh(+inf+i*2)",
			Complex.cosh(new Complex(POSINF,2)),
			new Complex(POSINF*Math.cos(2),POSINF*Math.sin(2)));
		check("cosh(+1+i*inf)",
			Complex.cosh(new Complex(1,POSINF)),
			new Complex(NaN,NaN));
	
		check("cosh(0+i*NaN)",
			Complex.cosh(new Complex(0,NaN)),
			new Complex(NaN,0));
		check("cosh(+inf+i*NaN)",
			Complex.cosh(new Complex(POSINF,NaN)),
			new Complex(POSINF,NaN));
		check("cosh(1+i*NaN)",
			Complex.cosh(new Complex(1,NaN)),
			new Complex(NaN,NaN));

		check("cosh(NaN+i*0)",
			Complex.cosh(new Complex(NaN,0)),
			new Complex(NaN,0));
		check("cosh(NaN+i)",
			Complex.cosh(new Complex(NaN,1)),
			new Complex(NaN,NaN));
		check("cosh(NaN+i*NaN)",
			Complex.cosh(new Complex(NaN,NaN)),
			new Complex(NaN,NaN));
	}

	void tanh()
	{
		check("tanh(2)", Complex.tanh(new Complex(2)), new Complex(Sfun.tanh(2)));

		// Special cases
		check("tanh(+0+i*0)",
			Complex.tanh(new Complex(0,0)),
			new Complex(0,0));
		check("tanh(+inf+i*2)",
			Complex.tanh(new Complex(POSINF,2)),
			new Complex(1,0));
		check("tanh(+1+i*inf)",
			Complex.tanh(new Complex(1,POSINF)),
			new Complex(NaN,NaN));
		check("tanh(+inf+i*NaN)",
			Complex.tanh(new Complex(POSINF,NaN)),
			new Complex(1,0));
		check("tanh(+NaN+i*0)",
			Complex.tanh(new Complex(NaN,0)),
			new Complex(NaN,0));
		check("tanh(+NaN+i*2)",
			Complex.tanh(new Complex(NaN,2)),
			new Complex(NaN,NaN));
		check("tanh(+2+i*NaN)",
			Complex.tanh(new Complex(2,NaN)),
			new Complex(NaN,NaN));
		check("tanh(+NaN+i*NaN)",
			Complex.tanh(new Complex(NaN,NaN)),
			new Complex(NaN,NaN));

		// Special cases (conjugate)
		check("tanh(+inf-i*2)",
			Complex.tanh(new Complex(POSINF,-2)),
			new Complex(1,0));
		check("tanh(+1-i*inf)",
			Complex.tanh(new Complex(1,NEGINF)),
			new Complex(NaN,NaN));
		check("tanh(+NaN-i*2)",
			Complex.tanh(new Complex(NaN,-2)),
			new Complex(NaN,NaN));

		// Special cases (oddness)
		check("tanh(-inf-i*2)",
			Complex.tanh(new Complex(NEGINF,2)),
			new Complex(-1,0));
		check("tanh(-1-i*inf)",
			Complex.tanh(new Complex(-1,NEGINF)),
			new Complex(NaN,NaN));
		check("tanh(-inf-i*NaN)",
			Complex.tanh(new Complex(NEGINF,NaN)),
			new Complex(-1,0));
		check("tanh(+NaN-i*2)",
			Complex.tanh(new Complex(NaN,-2)),
			new Complex(NaN,NaN));
		check("tanh(-2+i*NaN)",
			Complex.tanh(new Complex(-2,NaN)),
			new Complex(NaN,NaN));
	}

	void asinh()
	{
		Complex	z;

		z = new Complex(0.2,0.1);
		check(
			"sinh(asinh("+z+"))",
			Complex.sinh(Complex.asinh(z)),
			z);
		
		z = new Complex(1,2);
		check(
			"sinh(asinh("+z+"))",
			Complex.sinh(Complex.asinh(z)),
			z);

		for (int k = 0;  k < REPEAT;  k++) {
			z = new Complex(-2*4*random(),-2*4*random());
			check(
				"sinh(asinh("+z+"))",
				Complex.sinh(Complex.asinh(z)),
				z);
		}

		// Special cases
		check("asinh(+0+i*0)",
			Complex.asinh(new Complex(0,0)),
			new Complex(0,0));
		check("asinh(+inf+i*inf)",
			Complex.asinh(new Complex(POSINF,POSINF)),
			new Complex(POSINF,PI/4));
		check("asinh(2+i*inf)",
			Complex.asinh(new Complex(2,POSINF)),
			new Complex(POSINF,PI/2));
		check("asinh(+inf+i*2)",
			Complex.asinh(new Complex(POSINF,2)),
			new Complex(POSINF,0));
		check("asinh(NaN+i*inf)",
			Complex.asinh(new Complex(NaN,POSINF)),
			new Complex(POSINF,NaN));
		check("asinh(+inf+i*NaN)",
			Complex.asinh(new Complex(POSINF,NaN)),
			new Complex(POSINF,NaN));
		check("asinh(+NaN+i*0)",
			Complex.asinh(new Complex(NaN,0)),
			new Complex(NaN,0));
		check("asinh(+NaN+i*2)",
			Complex.asinh(new Complex(NaN,2)),
			new Complex(NaN,NaN));
		check("asinh(+2+i*NaN)",
			Complex.asinh(new Complex(2,NaN)),
			new Complex(NaN,NaN));
		check("asinh(NaN+i*NaN)",
			Complex.asinh(new Complex(NaN,NaN)),
			new Complex(NaN,NaN));

		// Special cases (conjugate)
		check("asinh(+0-i*0)",
			Complex.asinh(new Complex(0,NEGZERO)),
			new Complex(0,NEGZERO));
		check("asinh(+inf-i*inf)",
			Complex.asinh(new Complex(POSINF,NEGINF)),
			new Complex(POSINF,-PI/4));
		check("asinh(2-i*inf)",
			Complex.asinh(new Complex(2,NEGINF)),
			new Complex(POSINF,-PI/2));
		check("asinh(+inf-i*2)",
			Complex.asinh(new Complex(POSINF,-2)),
			new Complex(POSINF,NEGZERO));
		check("asinh(NaN-i*inf)",
			Complex.asinh(new Complex(NaN,NEGINF)),
			new Complex(NEGINF,NaN)); // sign of inf is indeterminant
		check("asinh(+NaN-i*0)",
			Complex.asinh(new Complex(NaN,NEGZERO)),
			new Complex(NaN,NEGZERO));
		check("asinh(+NaN-i*2)",
			Complex.asinh(new Complex(NaN,-2)),
			new Complex(NaN,NaN));

		// Special cases (oddness)
		check("asinh(-0-i*0)",
			Complex.asinh(new Complex(NEGZERO,NEGZERO)),
			new Complex(NEGZERO,NEGZERO));
		check("asinh(-inf-i*inf)",
			Complex.asinh(new Complex(NEGINF,NEGINF)),
			new Complex(NEGINF,-PI/4));
		check("asinh(-2-i*inf)",
			Complex.asinh(new Complex(-2,NEGINF)),
			new Complex(NEGINF,-PI/2));
		check("asinh(-inf-i*2)",
			Complex.asinh(new Complex(NEGINF,-2)),
			new Complex(NEGINF,NEGZERO));
		check("asinh(NaN-i*inf)",
			Complex.asinh(new Complex(NaN,NEGINF)),
			new Complex(NEGINF,NaN));
		check("asinh(-inf+i*NaN)",
			Complex.asinh(new Complex(NEGINF,NaN)),
			new Complex(NEGINF,NaN));
		check("asinh(+NaN-i*0)",
			Complex.asinh(new Complex(NaN,NEGZERO)),
			new Complex(NaN,NEGZERO));
		check("asinh(+NaN-i*2)",
			Complex.asinh(new Complex(NaN,-2)),
			new Complex(NaN,NaN));
		check("asinh(-2+i*NaN)",
			Complex.asinh(new Complex(-2,NaN)),
			new Complex(NaN,NaN));
	}

	void acosh()
	{
		Complex	z;

		z = new Complex(0.2,0.1);
		check(
			"cosh(acosh("+z+"))",
			Complex.cosh(Complex.acosh(z)),
			z);
		
		z = new Complex(1,2);
		check(
			"cosh(acosh("+z+"))",
			Complex.cosh(Complex.acosh(z)),
			z);

		for (int k = 0;  k < REPEAT;  k++) {
			z = new Complex(-2*4*random(),-2*4*random());
			check(
				"cosh(acosh("+z+"))",
				Complex.cosh(Complex.acosh(z)),
				z);
		}

		// Special cases
		check("acosh(+0+i*0)",
			Complex.acosh(new Complex(0,0)),
			new Complex(0,PI/2));
		check("acosh(-0+i*0)",
			Complex.acosh(new Complex(NEGZERO,0)),
			new Complex(0,PI/2));
		check("acosh(-inf+i*inf)",
			Complex.acosh(new Complex(NEGINF,POSINF)),
			new Complex(POSINF,0.75*PI));
		check("acosh(+inf+i*inf)",
			Complex.acosh(new Complex(POSINF,POSINF)),
			new Complex(POSINF,PI/4));
		check("acosh(2+i*inf)",
			Complex.acosh(new Complex(2,POSINF)),
			new Complex(POSINF,PI/2));
		check("acosh(-inf+i*2)",
			Complex.acosh(new Complex(NEGINF,2)),
			new Complex(POSINF,PI));
		check("acosh(+inf+i*2)",
			Complex.acosh(new Complex(POSINF,2)),
			new Complex(POSINF,0));
		check("acosh(+inf+i*NaN)",
			Complex.acosh(new Complex(POSINF,NaN)),
			new Complex(POSINF,NaN));
		check("acosh(-inf+i*NaN)",
			Complex.acosh(new Complex(NEGINF,NaN)),
			new Complex(POSINF,NaN));
		check("acosh(+2+i*NaN)",
			Complex.acosh(new Complex(2,NaN)),
			new Complex(NaN,NaN));
		check("acosh(+NaN+i*2)",
			Complex.acosh(new Complex(NaN,2)),
			new Complex(NaN,NaN));
		check("acosh(NaN+i*NaN)",
			Complex.acosh(new Complex(NaN,NaN)),
			new Complex(NaN,NaN));

		// Special cases (conjugate)
		check("acosh(+0-i*0)",
			Complex.acosh(new Complex(0,NEGZERO)),
			new Complex(0,-PI/2));
		check("acosh(-0-i*0)",
			Complex.acosh(new Complex(NEGZERO,NEGZERO)),
			new Complex(0,-PI/2));
		check("acosh(-inf-i*inf)",
			Complex.acosh(new Complex(NEGINF,NEGINF)),
			new Complex(POSINF,-0.75*PI));
		check("acosh(+inf-i*inf)",
			Complex.acosh(new Complex(POSINF,NEGINF)),
			new Complex(POSINF,-PI/4));
		check("acosh(2-i*inf)",
			Complex.acosh(new Complex(2,NEGINF)),
			new Complex(POSINF,-PI/2));
		check("acosh(-inf-i*2)",
			Complex.acosh(new Complex(NEGINF,-2)),
			new Complex(POSINF,-PI));
		check("acosh(+inf-i*2)",
			Complex.acosh(new Complex(POSINF,-2)),
			new Complex(POSINF,NEGZERO));
		check("acosh(+NaN-i*2)",
			Complex.acosh(new Complex(NaN,-2)),
			new Complex(NaN,NaN));
	}

	
	void atanh()
	{
		Complex	z;

		z = new Complex(0.2,0.1);
		check(
			"tanh(atanh("+z+"))",
			Complex.tanh(Complex.atanh(z)),
			z);
		
		z = new Complex(1,2);
		check(
			"tanh(atanh("+z+"))",
			Complex.tanh(Complex.atanh(z)),
			z);

		for (int k = 0;  k < REPEAT;  k++) {
			z = new Complex(-2*4*random(),-2*4*random());
			check(
				"tanh(atanh("+z+"))",
				Complex.tanh(Complex.atanh(z)),
				z);
		}

		// Special cases
		check("atanh(+0+i*0)",
			Complex.atanh(new Complex(0,0)),
			new Complex(0,0));
		check("atanh(+inf+i*inf)",
			Complex.atanh(new Complex(POSINF,POSINF)),
			new Complex(0,PI/2));
		check("atanh(+inf+i*2)",
			Complex.atanh(new Complex(POSINF,2)),
			new Complex(0,PI/2));
		check("atanh(+2+i*inf)",
			Complex.atanh(new Complex(2,POSINF)),
			new Complex(0,PI/2));
		check("atanh(+0+i*NaN)",
			Complex.atanh(new Complex(0,NaN)),
			new Complex(0,NaN));
		check("atanh(+NaN+i*inf)",
			Complex.atanh(new Complex(NaN,POSINF)),
			new Complex(NEGZERO,PI/2)); // real sign unspecified
		check("atanh(+inf+i*NaN)",
			Complex.atanh(new Complex(POSINF,NaN)),
			new Complex(0,NaN));
		check("atanh(+NaN+i*2)",
			Complex.atanh(new Complex(NaN,2)),
			new Complex(NaN,NaN));
		check("atanh(+2+i*NaN)",
			Complex.atanh(new Complex(2,NaN)),
			new Complex(NaN,NaN));
		check("atanh(+NaN+i*NaN)",
			Complex.atanh(new Complex(NaN,NaN)),
			new Complex(NaN,NaN));

		// Special cases (conjugate)
		check("atanh(+0-i*0)",
			Complex.atanh(new Complex(0,NEGZERO)),
			new Complex(0,NEGZERO));
		check("atanh(+inf-i*inf)",
			Complex.atanh(new Complex(POSINF,NEGINF)),
			new Complex(0,-PI/2));
		check("atanh(+inf-i*2)",
			Complex.atanh(new Complex(POSINF,-2)),
			new Complex(0,-PI/2));
		check("atanh(+2-i*inf)",
			Complex.atanh(new Complex(2,NEGINF)),
			new Complex(0,-PI/2));
		check("atanh(+NaN-i*inf)",
			Complex.atanh(new Complex(NaN,NEGINF)),
			new Complex(NEGZERO,-PI/2)); // sign unspecified
		check("atanh(+NaN-i*2)",
			Complex.atanh(new Complex(NaN,-2)),
			new Complex(NaN,NaN));

		// Special cases (oddness)
		check("atanh(-0-i*0)",
			Complex.atanh(new Complex(NEGZERO,NEGZERO)),
			new Complex(NEGZERO,NEGZERO));
		check("atanh(-inf-i*inf)",
			Complex.atanh(new Complex(NEGINF,NEGINF)),
			new Complex(NEGZERO,-PI/2));
		check("atanh(-inf-i*2)",
			Complex.atanh(new Complex(NEGINF,-2)),
			new Complex(NEGZERO,-PI/2));
		check("atanh(-2-i*inf)",
			Complex.atanh(new Complex(-2,NEGINF)),
			new Complex(NEGZERO,-PI/2));

		// Special cases (conjugate oddness)
		check("atanh(-0+i*0)",
			Complex.atanh(new Complex(NEGZERO,0)),
			new Complex(NEGZERO,0));
		check("atanh(-inf+i*inf)",
			Complex.atanh(new Complex(NEGINF,POSINF)),
			new Complex(NEGZERO,PI/2));
		check("atanh(-inf+i*2)",
			Complex.atanh(new Complex(NEGINF,2)),
			new Complex(NEGZERO,PI/2));
		check("atanh(-2+i*inf)",
			Complex.atanh(new Complex(-2,POSINF)),
			new Complex(NEGZERO,PI/2));
	}


	void pow()
	{
		Complex ans, z;
		
		z = new Complex(.2,.1);
		ans = Complex.times(z,z);
		check(
			"pow(0.2+i*0.1,2.0)",
			Complex.pow(z,2.0),
			ans);

		ans = Complex.times(Complex.sqrt(z),Complex.times(z,z));
		check(
			"pow(0.2+i*0.1,2.5)",
			Complex.pow(z,2.5),
			ans);
	}

	private void valueOfException(String input)
	{
		Complex ans = new Complex(999,999);
		Complex z   = new Complex();
		try {
			z = Complex.valueOf(input);
		} catch (NumberFormatException e) {
			// NumberFormatException is the right answer
			z = ans;
		}
		check("valueOf(NumberFormatException) "+input, z, ans);
	}

	void valueOf()
	{	
		check("valueOf 2+3i",  Complex.valueOf("2+3i"),  new Complex(2,3));
		check("valueOf -2+3i", Complex.valueOf("-2+3i"), new Complex(-2,3));
		check("valueOf 2-3i",  Complex.valueOf("2-3i"),  new Complex(2,-3));
		check("valueOf -2-3i", Complex.valueOf("-2-3i"), new Complex(-2,-3));

		check("valueOf 2+i",  Complex.valueOf("2+i"),  new Complex(2,1));
		check("valueOf -2+i", Complex.valueOf("-2+i"), new Complex(-2,1));
		check("valueOf 2-i",  Complex.valueOf("2-i"),  new Complex(2,-1));
		check("valueOf -2-i", Complex.valueOf("-2-i"), new Complex(-2,-1));

		check("valueOf 2",     Complex.valueOf("2"),     new Complex(2,0));
		check("valueOf i",     Complex.valueOf("i"),     new Complex(0,1));
		check("valueOf -i",    Complex.valueOf("-i"),    new Complex(0,-1));
		check("valueOf 3i",    Complex.valueOf("3i"),    new Complex(0,3));
		check("valueOf +3i",   Complex.valueOf("+3i"),   new Complex(0,3));
		check("valueOf -3i",   Complex.valueOf("-3i"),   new Complex(0,-3));

		check("valueOf 2.123-3.456i", Complex.valueOf("2.123-3.456i"), new Complex(2.123,-3.456));
		check("valueOf 2.123",		  Complex.valueOf("2.123"),        new Complex(2.123,0));
		check("valueOf -3.456i",      Complex.valueOf("-3.456i"),      new Complex(0,-3.456));
 
		check("valueOf .123-.456i",   Complex.valueOf(".123-.456i"),   new Complex(.123,-.456));
		check("valueOf .123",		  Complex.valueOf(".123"),         new Complex(.123,0));
		check("valueOf -.456i",       Complex.valueOf("-.456i"),       new Complex(0,-.456));

		check("valueOf -2.123e+3-3e4i", Complex.valueOf("-2.123e+3-3e4i"), new Complex(-2.123e+3,-3e4));
		check("valueOf -2.123e+3",		Complex.valueOf("-2.123e+3"),      new Complex(-2.123e+3,0));
		check("valueOf -3e4i",			Complex.valueOf("-3e4i"),		   new Complex(0,-3e4));

		try {
			check("valueOf null", Complex.valueOf(null), new Complex(0,0));
		} catch (NullPointerException e) {
		}
		
		valueOfException("2+");
		valueOfException("x");
		valueOfException("");
	}

	protected String getTestName()
	{
		return "com.imsl.Complex";
	}

	public void all()
	{
		setSeed(91323412);
		negative();
		conjugate();
		plus();
		minus();
		timesCC();
		timesDC();
		timesCD();
		divideCC();
		divideCD();
		abs();
		argument();
		sqrt();
		exp();
		log();
		sin();
		cos();
		tan();
		asin();
		acos();
		atan();
		sinh();
		cosh();
		tanh();
		asinh();
		acosh();
		atanh();
		pow();
		valueOf();
	}
}
