/*
 * =========================================================================
 * Copyright (C) 1997 - 1999 by Visual Numerics, Inc. All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software is freely
 * granted by Visual Numerics, Inc., provided that the copyright notice
 * above and the following warranty disclaimer are preserved in human
 * readable form.
 *
 * Because this software is licenses free of charge, it is provided
 * "AS IS", with NO WARRANTY.  TO THE EXTENT PERMITTED BY LAW, VNI
 * DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO ITS PERFORMANCE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * VNI WILL NOT BE LIABLE FOR ANY DAMAGES WHATSOEVER ARISING OUT OF THE USE
 * OF OR INABILITY TO USE THIS SOFTWARE, INCLUDING BUT NOT LIMITED TO DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, PUNITIVE, AND EXEMPLARY DAMAGES, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES. 
 *
 * =========================================================================
 */

package com.imsl.test.math;
import com.imsl.math.JMath;
import java.util.Random;

class TestJMath
{
	static final public double POSINF  = Double.POSITIVE_INFINITY;
	static final public double NEGINF  = Double.NEGATIVE_INFINITY;
	static final public double NaN     = Double.NaN;
	static final public double PI      = Math.PI;
	static final public double NEGZERO = 1/NEGINF;

	static public boolean logPassingTests = true;
	static public boolean logFailingTests = true;
	static private java.io.PrintStream out = System.out;
	static private int nTestsPassed = 0;
	static private int nTestsFailed = 0;

	private long seed;


	private void failHeader(String test)
	{
		out.println("Fail JMath." + test);
	}

	private boolean equals(double x, double y)
	{
		if (Double.isNaN(x) &&  Double.isNaN(y))  return true;
		return (Double.doubleToLongBits(x)==Double.doubleToLongBits(y));
	}

	final private static long negZeroBits = Double.doubleToLongBits(NEGZERO);

	static final private boolean isNegZero(double x)
	{
		return (Double.doubleToLongBits(x) == negZeroBits);
	}


	/**
	 *	Check the result of a test returning a double.
	 *	Test passes only if the results match exactly.
	 *	@param	test	Name of the test for the log file.
	 *	@param	compute The computed value.
	 *	@param	expect	The expected value.
	 */
	protected boolean check(String test, double compute, double expect)
	{
		long	Lcompute = Double.doubleToLongBits(compute);
		long	Lexpect  = Double.doubleToLongBits(expect);
		boolean pass = (Lcompute == Lexpect);

		if (pass) {
			if (logPassingTests)  out.println("Pass "+test);
			nTestsPassed++;
		} else {
			if (logFailingTests) {
				failHeader(test);		
				out.println("   Computed "+Long.toHexString(Lcompute)+"   "+compute);
				out.println("   Expected "+Long.toHexString(Lexpect)+"   "+expect);
				out.println("   "+ulp(compute,expect)+" ulp");
			}
			nTestsFailed++;
		}
		return pass;
	}


	/**
	 *	Return the difference between x and y in terms of units in the last place.
	 */
	long ulp(double x, double y)
	{
		boolean xIsNegative = (x < 0.0);
		boolean yIsNegative = (y < 0.0);

		long ix = Double.doubleToLongBits(x);
		long iy = Double.doubleToLongBits(y);

		int  mx = (int)((0x7FF0000000000000L&ix)>>52);	// biased exponent of x
		int  my = (int)((0x7FF0000000000000L&iy)>>52);  // biased exponent of y
		
		// Set ix and iy to the fractional part, with the
		// implicit bit added in.
		ix = (ix&0x000fffffffffffffL) | 0x0010000000000000L;
		iy = (iy&0x000fffffffffffffL) | 0x0010000000000000L;
		
		// Align fractions if the exponents differ
		if (mx > my) {
			ix <<= (mx-my);
		} else if (mx < my) {
			iy <<= (my-mx);
		}

		if (xIsNegative)  ix = -ix;
		if (yIsNegative)  iy = -iy;

		return (ix-iy);
	}


	void ceil()
	{
		check("ceil(0)",		JMath.ceil(0.0),		0.0);
		check("ceil(-0)",		JMath.ceil(NEGZERO),	NEGZERO);
		check("ceil(+inf)",		JMath.ceil(POSINF),		POSINF);
		check("ceil(-inf)",		JMath.ceil(NEGINF),		NEGINF);
		check("ceil(NaN)",		JMath.ceil(NaN),		NaN);

		check("ceil(0.99999)",	JMath.ceil(0.99999),	1.0);
		check("ceil(1)",		JMath.ceil(1.0),		1.0);
		check("ceil(1.00001)",	JMath.ceil(1.00001),	2.0);
		check("ceil(-0.99999)",	JMath.ceil(-0.99999),	NEGZERO);
		check("ceil(-1)",		JMath.ceil(-1.0),		-1.0);
		check("ceil(-1.00001)",	JMath.ceil(-1.00001),	-1.0);

		double x = 300292927534.71;
		check("ceil("+x+")",	JMath.ceil(x),			Math.ceil(x));
		check("ceil("+(-x)+")",	JMath.ceil(-x),			Math.ceil(-x));
		x = 300292927534.21;
		check("ceil("+x+")",	JMath.ceil(x),			Math.ceil(x));
		check("ceil("+(-x)+")",	JMath.ceil(-x),			Math.ceil(-x));
	}

	void floor()
	{
		check("floor(0)",		JMath.floor(0.0),		0.0);
		check("floor(-0)",		JMath.floor(NEGZERO),	NEGZERO);
		check("floor(inf)",		JMath.floor(POSINF),	POSINF);
		check("floor(-inf)",	JMath.floor(NEGINF),	NEGINF);
		check("floor(NaN)",		JMath.floor(NaN),		NaN);

		check("floor(0.99999)",	JMath.floor(0.99999),	0.0);
		check("floor(1)",		JMath.floor(1.0),		1.0);
		check("floor(1.00001)",	JMath.floor(1.00001),	1.0);
		check("floor(-0.99999)",JMath.floor(-0.99999),	-1.0);
		check("floor(-1)",		JMath.floor(-1.0),		-1.0);
		check("floor(-1.00001)",JMath.floor(-1.00001),	-2.0);

		double x = 300292927534.71;
		check("floor("+x+")",	JMath.floor(x),			Math.floor(x));
		check("floor("+(-x)+")",JMath.floor(-x),		Math.floor(-x));
		x = 300292927534.21;
		check("floor("+x+")",	JMath.floor(x),			Math.floor(x));
		check("floor("+(-x)+")",JMath.floor(-x),		Math.floor(-x));
	}
	
	void round()
	{
		check("round(0)",		JMath.round(0.0),		0.0);
		check("round(-0)",		JMath.round(NEGZERO),	0.0);
		check("round(+inf)",	JMath.round(POSINF),	Long.MAX_VALUE);
		check("round(-inf)",	JMath.round(NEGINF),	Long.MIN_VALUE);
		check("round(NaN)",		JMath.round(NaN),		0);

		check("round(0.99999)",	JMath.round(0.99999),	1.0);
		check("round(1)",		JMath.round(1.0),		1.0);
		check("round(1.00001)",	JMath.round(1.00001),	1.0);
		check("round(-0.99999)",JMath.round(-0.99999),	-1.0);
		check("round(-1)",		JMath.round(-1.0),		-1.0);
		check("round(-1.00001)",JMath.round(-1.00001),	-1.0);
	}

	void rint()
	{
		check("rint(0)",		JMath.rint(0.0),		0.0);
		check("rint(-0)",		JMath.rint(NEGZERO),	NEGZERO);
		check("rint(+inf)",		JMath.rint(POSINF),		POSINF);
		check("rint(-inf)",		JMath.rint(NEGINF),		NEGINF);
		check("rint(NaN)",		JMath.rint(NaN),		NaN);
		
		check("rint(0.99999)",	JMath.rint(0.99999),	1.0);
		check("rint(1)",		JMath.rint(1.0),		1.0);
		check("rint(1.00001)",	JMath.rint(1.00001),	1.0);
		check("rint(-0.99999)",	JMath.rint(-0.99999),	-1.0);
		check("rint(-1)",		JMath.rint(-1.0),		-1.0);
		check("rint(-1.00001)",	JMath.rint(-1.00001),	-1.0);

		double	x = -6.9646423920640664e16;
		check("rint("+x+")",	JMath.rint(x),			Math.rint(x));
		x = 1.7211153283167578e16;
		check("rint("+x+")",	JMath.rint(x),			Math.rint(x));
	}

	void abs()
	{
		check("abs(0)",			JMath.abs(0),		0);
		check("abs(-0)",		JMath.abs(NEGZERO),	0);
		check("abs(inf)",		JMath.abs(POSINF),	POSINF);
		check("abs(-inf)",		JMath.abs(NEGINF),	POSINF);
		check("abs(NaN)",		JMath.abs(NaN),		NaN);
		
		check("abs(16)",		JMath.abs(16),		16);
		check("abs(1e300)",		JMath.abs(1e300),	1e300);
		check("abs(-16)",		JMath.abs(-16),		16);
		check("abs(-1e300)",	JMath.abs(-1e300),	1e300);
	}
	
	void sqrt()
	{
		check("sqrt(0)",		JMath.sqrt(0),		0);
		check("sqrt(-0)",		JMath.sqrt(NEGZERO), NEGZERO);
		check("sqrt(-1)",		JMath.sqrt(-1),		NaN);
		check("sqrt(+inf)",		JMath.sqrt(POSINF),	POSINF);
		check("sqrt(-inf)",		JMath.sqrt(NEGINF),	NaN);
		check("sqrt(NaN)",		JMath.sqrt(NaN),	NaN);		

		check("sqrt(16)",		JMath.sqrt(16),		4);
		check("sqrt(1e300)",	JMath.sqrt(1e300),	1e150);
	}
	
	void exp()
	{
		check("exp(0)",			JMath.exp(0),		1);
		check("exp(-0)",		JMath.exp(NEGZERO),	1);
		check("exp(inf)",		JMath.exp(POSINF),	POSINF);
		check("exp(-inf)",		JMath.exp(NEGINF),	0);
		check("exp(NaN)",		JMath.exp(NaN),		NaN);

		check("exp(2.3)",		JMath.exp(2.3),		9.974182454814718);
		check("exp(2.3e-4)",	JMath.exp(2.3e-4),	1.0002300264520279);
		check("exp(2.3e2)",		JMath.exp(2.3e2),	7.722018499983836e99);

		long cases[] = { // breakpoints in the fdlibm code
			0x40862e42fefa39f0L,
			0x40862e42fefa39efL,
			0x40862e42fefa39eaL,
			0x7ff0000000000005L,
			0x7ff0000000000000L,
			0x7feffffffffffffaL,
			0x40862e4200000005L,
			0x40862e4200000000L,
			0x40862e41fffffffaL,
			0x3fd62e4200000005L,
			0x3fd62e4200000000L,
			0x3fd62e41fffffffaL,

			0x3ff0a2b200000005L,
			0x3ff0a2b200000000L,
			0x3ff0a2b1fffffffaL,
			0x3e30000000000005L,
			0x3e30000000000000L,
			0x3e2ffffffffffffaL,
		};
		for (int k = 0;  k < cases.length;  k++) {
			double x = Double.longBitsToDouble(cases[k]);
			check("exp("+x+")",  JMath.exp(x),    Math.exp(x));
			x = -1;
			check("exp("+x+")",  JMath.exp(x),    Math.exp(x));
		}
	}

	void log()
	{
		check("log(0)",			JMath.log(0),		NEGINF);
		check("log(-0)",		JMath.log(NEGZERO),	NEGINF);
		check("log(1)",			JMath.log(1),		0);
		check("log(-1)",		JMath.log(-1),		NaN);
		check("log(inf)",		JMath.log(POSINF),	POSINF);
		check("log(NaN)",		JMath.log(NaN),		NaN);
		
		check("log(100)",		JMath.log(100),		2*JMath.log(10));
		check("log(1e300)",		JMath.log(1e300),	Math.log(1e300));////300*JMath.log(10));
		check("log(-inf)",		JMath.log(NEGINF),	NaN);
	}

	void sin()
	{
		check("sin(0)",			JMath.sin(0),		0);
		check("sin(-0)",		JMath.sin(NEGZERO),	NEGZERO);
		check("sin(+inf)",		JMath.sin(POSINF),	NaN);
		check("sin(-inf)",		JMath.sin(NEGINF),	NaN);
		check("sin(NaN)",		JMath.sin(NaN),		NaN);

		check("sin(100*PI)"	,	JMath.sin(100*PI),	Math.sin(100*PI));
		check("sin(1e300*PI)",	JMath.sin(1e300*PI), Math.sin(1e300*PI));
	}
	
	void cos()
	{
		check("cos(0)",			JMath.cos(0),		1);
		check("cos(-0)",		JMath.cos(NEGZERO),	1);
		check("cos(+inf)",		JMath.cos(POSINF),	NaN);
		check("cos(-inf)",		JMath.cos(NEGINF),	NaN);
		check("cos(NaN)",		JMath.cos(NaN),		NaN);

		check("cos(100*PI)"	,	JMath.cos(100*PI),	1);
		check("cos(1e300*PI)",	JMath.cos(1e300*PI),Math.cos(1e300*PI));
	}

	void tan()
	{
		check("tan(0)",			JMath.tan(0),		0);
		check("tan(-0)",		JMath.tan(NEGZERO),	NEGZERO);
		check("tan(inf)",		JMath.tan(POSINF),	NaN);
		check("tan(-inf)",		JMath.tan(NEGINF),	NaN);
		check("tan(NaN)",		JMath.tan(NaN),		NaN);

		check("tan(100*PI)"	,	JMath.tan(100*PI),	Math.tan(100*PI));
		check("tan(1e300*PI)",	JMath.tan(1e300*PI),Math.tan(1e300*PI));
	}

	void asin()
	{
		check("asin(0)",		JMath.asin(0),		0);
		check("asin(-0)",		JMath.asin(NEGZERO),NEGZERO);
		check("asin(1)",		JMath.asin(1),		PI/2);
		check("asin(-1)",		JMath.asin(-1),		-PI/2);
		check("asin(2)",		JMath.asin(2),		NaN);
		check("asin(+inf)",		JMath.asin(POSINF),	NaN);
		check("asin(-inf)",		JMath.asin(NEGINF),	NaN);
		check("asin(NaN)",		JMath.asin(NaN),	NaN);
	}
	
	void acos()
	{
		check("acos(0)",		JMath.acos(0),		PI/2);
		check("acos(1)",		JMath.acos(1),		0);
		check("acos(2)",		JMath.acos(2),		NaN);
		check("acos(+inf)",		JMath.acos(POSINF),	NaN);
		check("acos(-inf)",		JMath.acos(NEGINF),	NaN);
		check("acos(-0)",		JMath.acos(NEGZERO),PI/2);
		check("acos(NaN)",		JMath.acos(NaN),	NaN);
	}

	void atan()
	{
		check("atan(0)",		JMath.atan(0),		0);
		check("atan(-0)",		JMath.atan(NEGZERO),NEGZERO);
		check("atan(1)",		JMath.atan(1),		PI/4);
		check("atan(-1)",		JMath.atan(-1),		-PI/4);
		check("atan(+inf)",		JMath.atan(POSINF),	PI/2);
		check("atan(-inf)",		JMath.atan(NEGINF),	-PI/2);
		check("atan(NaN)",		JMath.atan(NaN),	NaN);

		long cases[] = { // breakpoints in the fdlibm code
			0x7ff0000000000005L,
			0x7ff0000000000000L,
			0x7feffffffffffffaL,
			0x44fffffffffffffaL,
			0x4410000000000000L,
			0x4410000000000005L,
			0x3fdc000000000005L,
			0x3fdc000000000000L,
			0x3fdbfffffffffffaL,
			0x3e20000000000005L,
			0x3e20000000000000L,
			0x3e14444444444444L,
			0x3e1ffffffffffffaL,
			0x3ff3000000000005L,
			0x3ff3000000000000L,
			0x3ff2fffffffffffaL,
			0x3fe6000000000005L,
			0x3fe6000000000000L,
			0x3fe5fffffffffffaL,
			0x4003800000000005L,
			0x4003800000000000L,
			0x40037ffffffffffaL,
		};
		for (int k = 0;  k < cases.length;  k++) {
			double x = Double.longBitsToDouble(cases[k]);
			check("atan("+x+")",  JMath.atan(x),    Math.atan(x));
			x = -1;
			check("atan("+x+")",  JMath.atan(x),    Math.atan(x));
		}
	}

	void min()
	{
		check("min(0,0)",			JMath.min(0,0),			0);	
		check("min(0,1)",			JMath.min(0,1),			0);
		check("min(1,0)",			JMath.min(1,0),			0);
		check("min(1,1)",			JMath.min(1,1),			1);

		check("min(0,-1)",			JMath.min(0,-1),		-1);
		check("min(-0,-1)",			JMath.min(NEGZERO,-1),	-1);
		check("min(-1,0)",			JMath.min(-1,0),		-1);
		check("min(-1,-1)",			JMath.min(-1,-1),		-1);

		check("min(-0,0)",			JMath.min(NEGZERO,0),	NEGZERO);
		check("min(-0,1)",			JMath.min(NEGZERO,1),	NEGZERO);
		check("min(0,-0)",			JMath.min(0, NEGZERO),	NEGZERO);
		check("min(0,-1)",			JMath.min(1, NEGZERO),	NEGZERO);
		
		check("min(0,+inf)",		JMath.min(0,POSINF),	0);
		check("min(0,-inf)",		JMath.min(0,NEGINF),	NEGINF);
		check("min(1,+inf)",		JMath.min(1,POSINF),	1);
		check("min(1,-inf)",		JMath.min(1,NEGINF),	NEGINF);

		check("min(+inf,0)",		JMath.min(POSINF,0),	0);
		check("min(-inf,0)",		JMath.min(NEGINF,0),	NEGINF);
		check("min(+inf,1)",		JMath.min(POSINF,1),	1);
		check("min(-inf,1)",		JMath.min(NEGINF,1),	NEGINF);

		check("min(+inf,+inf)",		JMath.min(POSINF,POSINF),	POSINF);
		check("min(+inf,-inf)",		JMath.min(POSINF,NEGINF),	NEGINF);
		check("min(-inf,+inf)",		JMath.min(NEGINF,POSINF),	NEGINF);
		check("min(-inf,-inf)",		JMath.min(NEGINF,NEGINF),	NEGINF);
		
		check("min(0,NaN)",			JMath.min(0,NaN),		NaN);
		check("min(NaN,0)",			JMath.min(NaN,0),		NaN);
		check("min(1,NaN)",			JMath.min(1,NaN),		NaN);
		check("min(NaN,1)",			JMath.min(NaN,1),		NaN);
		check("min(+inf,NaN)",		JMath.min(POSINF,NaN),	NaN);
		check("min(NaN,+inf)",		JMath.min(NaN,POSINF),	NaN);
	}

	void max()
	{
		check("max(0,0)",			JMath.max(0,0),			0);	
		check("max(0,1)",			JMath.max(0,1),			1);
		check("max(1,0)",			JMath.max(1,0),			1);
		check("max(1,1)",			JMath.max(1,1),			1);

		check("max(0,-1)",			JMath.max(0,-1),		0);
		check("max(-0,-1)",			JMath.max(NEGZERO,-1),	NEGZERO);
		check("max(-1,0)",			JMath.max(-1,0),		0);
		check("max(-1,-1)",			JMath.max(-1,-1),		-1);

		check("max(-0,0)",			JMath.max(NEGZERO,0),	0);
		check("max(-0,1)",			JMath.max(NEGZERO,1),	1);
		check("max(0,-0)",			JMath.max(0,NEGZERO),	0);
		check("max(0,-1)",			JMath.max(1,NEGZERO),	1);
		
		check("max(0,+inf)",		JMath.max(0,POSINF),	POSINF);
		check("max(0,-inf)",		JMath.max(0,NEGINF),	0);
		check("max(1,+inf)",		JMath.max(1,POSINF),	POSINF);
		check("max(1,-inf)",		JMath.max(1,NEGINF),	1);

		check("max(+inf,0)",		JMath.max(POSINF,0),	POSINF);
		check("max(-inf,0)",		JMath.max(NEGINF,0),	0);
		check("max(+inf,1)",		JMath.max(POSINF,1),	POSINF);
		check("max(-inf,1)",		JMath.max(NEGINF,1),	1);

		check("max(+inf,+inf)",		JMath.max(POSINF,POSINF),POSINF);
		check("max(+inf,-inf)",		JMath.max(POSINF,NEGINF),POSINF);
		check("max(-inf,+inf)",		JMath.max(NEGINF,POSINF),POSINF);
		check("max(-inf,-inf)",		JMath.max(NEGINF,NEGINF),NEGINF);
		
		check("max(0,NaN)",			JMath.max(0,NaN),		NaN);
		check("max(NaN,0)",			JMath.max(NaN,0),		NaN);
		check("max(1,NaN)",			JMath.max(1,NaN),		NaN);
		check("max(NaN,1)",			JMath.max(NaN,1),		NaN);
		check("max(+inf,NaN)",		JMath.max(POSINF,NaN),	NaN);
		check("max(NaN,+inf)",		JMath.max(NaN,POSINF),	NaN);
	}

	void atan2()
	{
		check("atan2(0,1)",			JMath.atan2(0,1),		0);
		check("atan2(-0,1)",		JMath.atan2(NEGZERO,1),	NEGZERO);
		
		check("atan2(0,0)",			JMath.atan2(0,0),		0);
		check("atan2(-0,0)",		JMath.atan2(NEGZERO,0),	NEGZERO);

		check("atan2(0,-0)",		JMath.atan2(0,NEGZERO),	PI);
		check("atan2(-0,-0)",		JMath.atan2(NEGZERO,NEGZERO),-PI);

		check("atan2(1,0)",			JMath.atan2(1,0),		PI/2);
		check("atan2(1,-0)",		JMath.atan2(1,NEGZERO),	PI/2);
		
		check("atan2(-1,0)",		JMath.atan2(-1,0),		-PI/2);
		check("atan2(-1,-0)",		JMath.atan2(-1,NEGZERO),	-PI/2);

		check("atan2(1,+inf)",		JMath.atan2(1,POSINF),	0);
		check("atan2(-1,+inf)",		JMath.atan2(-1,POSINF),	NEGZERO);

		check("atan2(+inf,1)",		JMath.atan2(POSINF,1),	PI/2);
		check("atan2(-inf,1)",		JMath.atan2(NEGINF,1),	-PI/2);
		
		check("atan2(1,-inf)",		JMath.atan2(1,NEGINF),	PI);
		check("atan2(-1,-inf)",		JMath.atan2(-1,NEGINF),	-PI);

		check("atan2(+inf,+inf)",	JMath.atan2(POSINF,POSINF),	PI/4);
		check("atan2(-inf,+inf)",	JMath.atan2(NEGINF,POSINF),	-PI/4);

		check("atan2(+inf,-inf)",	JMath.atan2(POSINF,NEGINF),	3*PI/4);
		check("atan2(-inf,-inf)",	JMath.atan2(NEGINF,NEGINF),	-3*PI/4);
		
		check("atan2(1,1)",			JMath.atan2(1,1),		PI/4);

		check("atan2(0,-1)",		JMath.atan2(0,-1),		PI);
		check("atan2(-1,0)",		JMath.atan2(-1,0),		-PI/2);
		check("atan2(-1,-1)",		JMath.atan2(-1,-1),		-0.75*PI);

		check("atan2(1e300,1e300)",		JMath.atan2(1e300,1e300),	PI/4);
		check("atan2(1e300,-1e300)",	JMath.atan2(1e300,-1e300),	3*PI/4);
		check("atan2(-1e300,1e300)",	JMath.atan2(-1e300,1e300),	-PI/4);
		check("atan2(-1e300,-1e300)",	JMath.atan2(-1e300,-1e300),	-3*PI/4);
		
		check("atan2(0,+inf)",		JMath.atan2(0,POSINF),	0);
		check("atan2(0,-inf)",		JMath.atan2(0,NEGINF),	PI);

		check("atan2(+inf,0)",		JMath.atan2(POSINF,0),	PI/2);
		check("atan2(-inf,0)",		JMath.atan2(NEGINF,0),	-PI/2);

		check("atan2(-inf,-inf)",	JMath.atan2(NEGINF,NEGINF),	-3*PI/4);
		
		check("atan2(0,NaN)",		JMath.atan2(0,NaN),		NaN);
		check("atan2(NaN,0)",		JMath.atan2(NaN,0),		NaN);
		check("atan2(1,NaN)",		JMath.atan2(1,NaN),		NaN);
		check("atan2(NaN,1)",		JMath.atan2(NaN,1),		NaN);
		check("atan2(+inf,NaN)",	JMath.atan2(POSINF,NaN),	NaN);
		check("atan2(NaN,+inf)",	JMath.atan2(NaN,POSINF),	NaN);

		Random r = new Random(seed^023556734L);
		for (int k = 0;  k < 0;  k++) {
			double x = 100*r.nextGaussian();
			double y = 100*r.nextGaussian();
			check("atan2("+x+","+y+")",
				JMath.atan2(x,y),
				Math.atan2(x,y));
		}
	}

	void pow()
	{
		check("pow(0,0)",			JMath.pow(0,0),			1);
		check("pow(-0,0)",			JMath.pow(NEGZERO,0),	1);
		check("pow(1,0)",			JMath.pow(1,0),			1);
		check("pow(-1,0)",			JMath.pow(-1,0),		1);
		check("pow(+inf,0)",		JMath.pow(POSINF,0),	1);
		check("pow(-inf,0)",		JMath.pow(NEGINF,0),	1);
		check("pow(NaN,0)",			JMath.pow(NaN,0),		1);

		check("pow(1.1,+inf)",		JMath.pow(1.1,POSINF),	POSINF);
		check("pow(0.9,+inf)",		JMath.pow(0.9,POSINF),	0);
		check("pow(0,+inf)",		JMath.pow(0,POSINF),	0);
		check("pow(-1.1,+inf)",		JMath.pow(-1.1,POSINF),	POSINF);
		check("pow(-0.9,+inf)",		JMath.pow(-0.9,POSINF),	0);

		check("pow(1.1,-inf)",		JMath.pow(1.1,NEGINF),	0);
		check("pow(0.9,-inf)",		JMath.pow(0.9,NEGINF),	POSINF);
		check("pow(0,-inf)",		JMath.pow(0,NEGINF),	POSINF);
		check("pow(-1.1,-inf)",		JMath.pow(-1.1,NEGINF),	0);
		check("pow(-0.9,-inf)",		JMath.pow(-0.9,NEGINF),	POSINF);

		check("pow(+inf,1)",		JMath.pow(POSINF,1),	POSINF);
		check("pow(+inf,+inf)",		JMath.pow(POSINF,POSINF),POSINF);
		check("pow(+inf,-1)",		JMath.pow(POSINF,-1),	0);
		check("pow(+inf,-inf)",		JMath.pow(POSINF,NEGINF),0);

		check("pow(-inf,1)",		JMath.pow(NEGINF,1.0),	NEGINF);
		check("pow(-inf,3)",		JMath.pow(NEGINF,3.0),	NEGINF);
		check("pow(-inf,2)",		JMath.pow(NEGINF,2.0),	POSINF);
		check("pow(-inf,2.5)",		JMath.pow(NEGINF,2.5),	POSINF);
		check("pow(-inf,+inf)",		JMath.pow(NEGINF,POSINF),POSINF);
		check("pow(-inf,-3)",		JMath.pow(NEGINF,-3),	NEGZERO);
		check("pow(-inf,-2.5)",		JMath.pow(NEGINF,-2.5),	0);
		check("pow(-inf,-inf)",		JMath.pow(NEGINF,NEGINF),0);

		check("pow(0,NaN)",			JMath.pow(0,NaN),		NaN);
		check("pow(1,NaN)",			JMath.pow(1,NaN),		NaN);
		check("pow(+inf,NaN)",		JMath.pow(POSINF,NaN),	NaN);
		check("pow(-inf,NaN)",		JMath.pow(NEGINF,NaN),	NaN);
		
		
		check("pow(NaN,1)",			JMath.pow(NaN,1),		NaN);
		check("pow(NaN,+inf)",		JMath.pow(NaN,POSINF),	NaN);
		
		check("pow(1,+inf)",		JMath.pow(1,POSINF),	NaN);
		check("pow(-1,+inf)",		JMath.pow(-1,POSINF),	NaN);
		check("pow(1,-inf)",		JMath.pow(1,NEGINF),	NaN);
		check("pow(-1,-inf)",		JMath.pow(-1,NEGINF),	NaN);

		check("pow(-2,2.5)",		JMath.pow(-2,2.5),		NaN);
		
		check("pow(0,-3)",			JMath.pow(0,-3),		POSINF);
		check("pow(-0,-3)",			JMath.pow(NEGZERO,-3),	NEGINF);

		check("pow(0,-2.5)",		JMath.pow(0,-2.5),		POSINF);
		check("pow(-0,-2.5)",		JMath.pow(NEGZERO,-2.5),POSINF);

		check("pow(0,3)",			JMath.pow(0,3),			0);
		check("pow(-0,3)",			JMath.pow(NEGZERO,3),	NEGZERO);
		
		check("pow(0,2.5)",			JMath.pow(0,2.5),		0);
		check("pow(-0,2.5)",		JMath.pow(NEGZERO,2.5),	0);


		check("pow(1,1)",			JMath.pow(1,1),			1);
		check("pow(0,-1)",			JMath.pow(0,-1),		POSINF);
		check("pow(-0,-1)",			JMath.pow(NEGZERO,-1),	NEGINF);
		check("pow(-1,-1)",			JMath.pow(-1,-1),		-1);
		check("pow(-0,1)",			JMath.pow(NEGZERO,1),	NEGZERO);
		check("pow(-0,2)",			JMath.pow(NEGZERO,2),	0);

		check("pow(-3.125,-3)",		JMath.pow(-3.125,-3.0),	-0.032768);

		check("pow(1e5,23)",		JMath.pow(1e5,23),		1e115);

		check(
			"pow(4.515629286686633E-13,25.836293630090577)",
			JMath.pow(4.515629286686633e-13,25.836293630090577),
			Math.pow(4.515629286686633e-13,25.836293630090577));
	}

	void IEEEremainder()
	{
		check("IEEEremainder(0,0)",		JMath.IEEEremainder(0,0),			NaN);
		check("IEEEremainder(0,-0)",	JMath.IEEEremainder(0,NEGZERO),		NaN);
		check("IEEEremainder(-0,0)",	JMath.IEEEremainder(NEGZERO,0),		NaN);
		check("IEEEremainder(-0,-0)",	JMath.IEEEremainder(NEGZERO,NEGZERO),NaN);

		check("IEEEremainder(1,0)",		JMath.IEEEremainder(1,0),			NaN);
		check("IEEEremainder(-1,0)",	JMath.IEEEremainder(-1,0),			NaN);
		check("IEEEremainder(1,-0)",	JMath.IEEEremainder(1,NEGZERO),		NaN);
		check("IEEEremainder(-1,-0)",	JMath.IEEEremainder(-1,NEGZERO),	NaN);
		
		check("IEEEremainder(0,1)",		JMath.IEEEremainder(0,1),			0);
		check("IEEEremainder(0,-1)",	JMath.IEEEremainder(0,-1),			0);
		
		check("IEEEremainder(-0,1)",	JMath.IEEEremainder(NEGZERO,1),		NEGZERO);
		check("IEEEremainder(-0,-1)",	JMath.IEEEremainder(NEGZERO,-1),	NEGZERO);

		check("IEEEremainder(5,1)",		JMath.IEEEremainder(5,1),			0);
		check("IEEEremainder(5,-1)",	JMath.IEEEremainder(5,-1),			0);

		check("IEEEremainder(-5,1)",	JMath.IEEEremainder(-5,1),			NEGZERO);
		check("IEEEremainder(-5,-1)",	JMath.IEEEremainder(-5,-1),			NEGZERO);

		
		check("IEEEremainder(+inf,+inf)",	JMath.IEEEremainder(POSINF,POSINF),	NaN);
		check("IEEEremainder(+inf,-inf)",	JMath.IEEEremainder(POSINF,NEGINF),	NaN);
		check("IEEEremainder(-inf,+inf)",	JMath.IEEEremainder(NEGINF,POSINF),	NaN);
		check("IEEEremainder(-inf,-inf)",	JMath.IEEEremainder(NEGINF,NEGINF),	NaN);

		check("IEEEremainder(+inf,0)",		JMath.IEEEremainder(POSINF,0),		NaN);
		check("IEEEremainder(-inf,0)",		JMath.IEEEremainder(NEGINF,0),		NaN);
		check("IEEEremainder(+inf,1)",		JMath.IEEEremainder(POSINF,1),		NaN);
		check("IEEEremainder(-inf,1)",		JMath.IEEEremainder(NEGINF,1),		NaN);
		check("IEEEremainder(+inf,-1)",		JMath.IEEEremainder(POSINF,-1),		NaN);
		check("IEEEremainder(-inf,-1)",		JMath.IEEEremainder(NEGINF,-1),		NaN);

		check("IEEEremainder(0,+inf)",		JMath.IEEEremainder(0,POSINF),		0);
		check("IEEEremainder(0,-inf)",		JMath.IEEEremainder(0,NEGINF),		0);
		check("IEEEremainder(1,+inf)",		JMath.IEEEremainder(1,POSINF),		1);
		check("IEEEremainder(1,-inf)",		JMath.IEEEremainder(1,NEGINF),		1);
		check("IEEEremainder(-1,inf)",		JMath.IEEEremainder(-1,POSINF),		-1);
		check("IEEEremainder(-1,-inf)",		JMath.IEEEremainder(-1,NEGINF),		-1);

		check("IEEEremainder(0,NaN)",		JMath.IEEEremainder(0,NaN),			NaN);
		check("IEEEremainder(NaN,0)",		JMath.IEEEremainder(NaN,0),			NaN);
		check("IEEEremainder(1,NaN)",		JMath.IEEEremainder(1,NaN),			NaN);
		check("IEEEremainder(NaN,1)",		JMath.IEEEremainder(NaN,1),			NaN);
		check("IEEEremainder(+inf,NaN)",	JMath.IEEEremainder(POSINF,NaN),	NaN);
		check("IEEEremainder(NaN,+inf)",	JMath.IEEEremainder(NaN,POSINF),	NaN);
	}


	void atIntegers(int nTrials)
	{
		double y = 0.0;
		for (int k = -nTrials; k < nTrials;  k++) {
			double x = k;
			check("ceil("+x+")",  JMath.ceil(x),  Math.ceil(x));
			check("floor("+x+")", JMath.floor(x), Math.floor(x));
			check("rint("+x+")",  JMath.rint(x),  Math.rint(x));
			check("sqrt("+x+")",  JMath.sqrt(x),  Math.sqrt(x));
			check("exp("+x+")",   JMath.exp(x),   Math.exp(x));
			check("log("+x+")",   JMath.log(x),	  Math.log(x));
			check("sin("+x+")",   JMath.sin(x),	  Math.sin(x));
			check("cos("+x+")",   JMath.cos(x),	  Math.cos(x));
			check("tan("+x+")",   JMath.tan(x),	  Math.tan(x));		
			double xs = x / nTrials;
			check("asin("+xs+")", JMath.asin(xs), Math.asin(xs));
			check("acos("+xs+")", JMath.acos(xs), Math.acos(xs));
			check("atan("+x+")",  JMath.atan(x),  Math.atan(x));
		}

		int		n = (int)Math.sqrt(nTrials);
		for (int j = -n; j < n;  j++) {
			double x = j;
			for (int k = -n; k < n;  k++) {
				y = k;
				check("atan2("+x+","+y+")",  JMath.atan2(x,y),  Math.atan2(x,y));
				check("pow("+x+","+y+")",    JMath.pow(x,y),    Math.pow(x,y));
				check("IEEEremainder("+x+","+y+")", JMath.IEEEremainder(x,y), Math.IEEEremainder(x,y));
			}
		}
	}


	void atGaussian(int nTrials, double mean, double sd)
	{
		double	y = 0.0;
		Random r = new Random(seed^0x123457L);
		for (int k = 0;  k < nTrials;  k++) {
			double x = mean + sd*r.nextGaussian();
			check("ceil("+x+")",  JMath.ceil(x),  Math.ceil(x));
			check("floor("+x+")", JMath.floor(x), Math.floor(x));
			check("rint("+x+")",  JMath.rint(x),  Math.rint(x));
			check("sqrt("+x+")",  JMath.sqrt(x),  Math.sqrt(x));
			check("exp("+x+")",   JMath.exp(x),	  Math.exp(x));
			check("log("+x+")",   JMath.log(x),	  Math.log(x));
			check("sin("+x+")",   JMath.sin(x),	  Math.sin(x));
			check("cos("+x+")",   JMath.cos(x),	  Math.cos(x));
			check("tan("+x+")",   JMath.tan(x),	  Math.tan(x));
			double xs = x;
			check("asin("+xs+")", JMath.asin(xs), Math.asin(xs));
			check("acos("+xs+")", JMath.acos(xs), Math.acos(xs));
			check("atan("+x+")",  JMath.atan(x),  Math.atan(x));
			check("atan2("+x+","+y+")",  JMath.atan2(x,y),  Math.atan2(x,y));
			check("pow("+x+","+y+")",    JMath.pow(x,y),    Math.pow(x,y));
			check("IEEEremainder("+x+","+y+")", JMath.IEEEremainder(x,y), Math.IEEEremainder(x,y));
			y = x;
		}
	}


	void atSubnormal(int nTrials)
	{
		double	y = 0.0;
		Random r = new Random(seed^0x893a5bL);
		for (int k = 0;  k < nTrials;  k++) {
			long lx = 0L;
			if (k % 5 == 1) {
				// for some k's, use "small" subnormal
				lx = (long)r.nextInt();
			} else {
				// form most k's use "large" subnormals
				lx = r.nextLong() >>> 12;
			}
			// for k odd make x < 0
			if (k % 2 == 1) lx |= 0x8000000000000000L;
			double x = Double.longBitsToDouble(lx);
			check("ceil("+x+")",  JMath.ceil(x),  Math.ceil(x));
			check("floor("+x+")", JMath.floor(x), Math.floor(x));
			check("rint("+x+")",  JMath.rint(x),  Math.rint(x));
			check("sqrt("+x+")",  JMath.sqrt(x),  Math.sqrt(x));
			check("exp("+x+")",   JMath.exp(x),	  Math.exp(x));
			check("log("+x+")",   JMath.log(x),	  Math.log(x));
			check("sin("+x+")",   JMath.sin(x),	  Math.sin(x));
			check("cos("+x+")",   JMath.cos(x),	  Math.cos(x));
			check("tan("+x+")",   JMath.tan(x),	  Math.tan(x));
			check("asin("+x+")",  JMath.asin(x),  Math.asin(x));
			check("acos("+x+")",  JMath.acos(x),  Math.acos(x));
			check("atan("+x+")",  JMath.atan(x),  Math.atan(x));
			check("atan2("+x+","+y+")",  JMath.atan2(x,y),  Math.atan2(x,y));
			check("pow("+x+","+y+")",    JMath.pow(x,y),    Math.pow(x,y));
			check("IEEEremainder("+x+","+y+")", JMath.IEEEremainder(x,y), Math.IEEEremainder(x,y));
			y = x;
		}
	}


	/*
	 *	Test numbers with all possible exponents
	 */
	void atExponents()
	{
		double	y = 0.0;
		for (long lx = 0L;  lx <= 0xFFFF;  lx++) {
			double x = Double.longBitsToDouble(lx<<(64-16));
			check("ceil("+x+")",  JMath.ceil(x),  Math.ceil(x));
			check("floor("+x+")", JMath.floor(x), Math.floor(x));
			check("rint("+x+")",  JMath.rint(x),  Math.rint(x));
			check("sqrt("+x+")",  JMath.sqrt(x),  Math.sqrt(x));
			check("exp("+x+")",   JMath.exp(x),	  Math.exp(x));
			check("log("+x+")",   JMath.log(x),	  Math.log(x));
			check("sin("+x+")",   JMath.sin(x),	  Math.sin(x));
			check("cos("+x+")",   JMath.cos(x),	  Math.cos(x));
			check("tan("+x+")",   JMath.tan(x),	  Math.tan(x));
			check("asin("+x+")",  JMath.asin(x),  Math.asin(x));
			check("acos("+x+")",  JMath.acos(x),  Math.acos(x));
			check("atan("+x+")",  JMath.atan(x),  Math.atan(x));
			check("atan2("+x+","+y+")",  JMath.atan2(x,y),  Math.atan2(x,y));
			check("pow("+x+","+y+")",    JMath.pow(x,y),    Math.pow(x,y));
			check("IEEEremainder("+x+","+y+")", JMath.IEEEremainder(x,y), Math.IEEEremainder(x,y));
			y = x;
		}
	}


	/*
	 *	Test functions of two arguments with
	 *	(integer,random) and (random,integer) values.
	 */
	void atMixed(int nTrials)
	{
		int		n = (int)Math.sqrt(nTrials);
		double	y = 0.0;
		Random r = new Random(seed^0x91a67dL);
		for (int iy = -n;  iy < n;  iy++) {
			y = iy;
			for (int k = 0;  k < n;  k++) {
				long ix = r.nextLong();
				double x = Double.longBitsToDouble(ix);
				check("atan2("+x+","+y+")",
					JMath.atan2(x,y),
					Math.atan2(x,y));
				check("atan2("+y+","+x+")",
					JMath.atan2(y,x),
					Math.atan2(y,x));
				check("pow("+x+","+y+")", 
					JMath.pow(x,y),
					Math.pow(x,y));
				check("pow("+y+","+x+")", 
					JMath.pow(y,x),
					Math.pow(y,x));
				check("IEEEremainder("+x+","+y+")",
					JMath.IEEEremainder(x,y),
					Math.IEEEremainder(x,y));
				check("IEEEremainder("+y+","+x+")",
					JMath.IEEEremainder(y,x),
					Math.IEEEremainder(y,x));
			}
		}
	}
	
	
	void atRandom(int nTrials)
	{
		double	y = 0.0;
		Random r = new Random(seed^0x91a67dL);
		for (int k = 0;  k < nTrials;  k++) {
			long ix = r.nextLong();
			double x = Double.longBitsToDouble(ix);
			check("ceil("+x+")",  JMath.ceil(x),  Math.ceil(x));
			check("floor("+x+")", JMath.floor(x), Math.floor(x));
			check("rint("+x+")",  JMath.rint(x),  Math.rint(x));
			check("sqrt("+x+")",  JMath.sqrt(x),  Math.sqrt(x));
			check("exp("+x+")",   JMath.exp(x),	  Math.exp(x));
			check("log("+x+")",   JMath.log(x),	  Math.log(x));
			check("sin("+x+")",   JMath.sin(x),	  Math.sin(x));
			check("cos("+x+")",   JMath.cos(x),	  Math.cos(x));
			check("tan("+x+")",   JMath.tan(x),	  Math.tan(x));
			check("asin("+x+")",  JMath.asin(x),  Math.asin(x));
			check("acos("+x+")",  JMath.acos(x),  Math.acos(x));
			check("atan("+x+")",  JMath.atan(x),  Math.atan(x));
			check("atan2("+x+","+y+")",  JMath.atan2(x,y),  Math.atan2(x,y));
			check("pow("+x+","+y+")",    JMath.pow(x,y),    Math.pow(x,y));
			check("IEEEremainder("+x+","+y+")", JMath.IEEEremainder(x,y), Math.IEEEremainder(x,y));
			y = x;
		}
	}
	

	void runAllTests(int nTrials, long seed)
	{
		this.seed = ((seed==0L) ?  System.currentTimeMillis() : seed);

		check("PI", JMath.PI, Math.PI);
		check("E",  JMath.E,  Math.E);
		ceil();
		floor();
		round();
		rint();
		abs();
		sqrt();
		exp();
		log();
		sin();
		cos();
		tan();
		asin();
		acos();
		atan();
		min();
		max();
		atan2();
		pow();
		IEEEremainder();
		atMixed(nTrials);
		atExponents();
		atRandom(nTrials);
		atSubnormal(nTrials);
		atIntegers(nTrials);
		atGaussian(nTrials, 0.0, 1.0);
		atGaussian(nTrials, 0.0, 100.0);
		atGaussian(nTrials, 500.0, 100.0);
		atGaussian(nTrials, -200.0, 100.0);
		atGaussian(nTrials, 0.0, 1.0e100);
	}


	public static void main(String args[])
	{
		System.out.println("Vendor:  "+System.getProperties().getProperty("java.vendor"));
		System.out.println("Version: "+System.getProperties().getProperty("java.version"));
		System.out.println("OS Name: "+System.getProperties().getProperty("os.name"));
		System.out.println("OS Arch: "+System.getProperties().getProperty("os.arch"));
		System.out.println("OS Vers: "+System.getProperties().getProperty("os.version"));
		
		logPassingTests = false;
		logFailingTests = true;

		int nTrials = 100;
		if (args.length > 0) {
			nTrials = Integer.valueOf(args[0]).intValue();
		}

		try {
			new TestJMath().runAllTests(nTrials, 0L);
		} catch (Exception exception) {
			exception.printStackTrace();
		}
		System.out.println("Test passing " + nTestsPassed);
		System.out.println("Test failing " + nTestsFailed);
	}
}
